/// <reference path="lodash.d.ts" />
/// <reference path="three.d.ts" />
/// <reference path="Utils.ts" />
/// <reference path="OrbitalMechanics.ts" />
/// <reference path="Rings.ts" />
/// <reference path="SOITree.ts" />
/// <reference path="SolarSystem.ts" />
/// <reference path="Labels.ts" />

// Body.ts

class Body {
  public radius : number;
  public color : Color;
  public material = null as THREE.Material;

  public orbitData : OrbitData;
  public rotationData : RotationData;

  public orbit : Orbit;
  public rotation : Rotation;

  constructor(
    public name : string,
    public data : BodyData,
    public kind = "planet"
  ) {
    this.radius = data.radius;
    this.orbitData = data.orbit;
    this.rotationData = data.rotation;
    this.color = HexCol(data.color);
    this.material = data.material || null;

    this.orbit = this.orbitData ? new Orbit(this.orbitData) : null;
    this.rotation = this.rotationData ? new Rotation(this.rotationData) : null;
  }

  toString() : string {
    return "[Body name=" + this.name + ", radius=" + (this.radius * 1e-3).toFixed(2) + "km]";
  }
}

/**
 * Responsible for rendering one SOI level.
 */
class LevelRenderer {
  private _setUp : boolean;
  constructor(
    public node : SOINode,
    public treeRenderer : SOITreeRenderer,
    public scene = treeRenderer.scene
  ) {
    console.log("  --- new LevelRenderer");
    this._setUp = false;
  }

  setUp() {
    if (this._setUp)
      console.error("setUp called twice!", this);

    console.log("    {{{ LevelRenderer::setUp", this.node.data);

    this.getNodes().forEach((node) => {
      if (node.storage._nodeRenderer) {
        console.info("This node already have a central mesh.", this, node);
        return;
      }

      var nodeRenderer = this.makeNodeRenderer(node);
      nodeRenderer.setUp();
      node.storage._nodeRenderer = nodeRenderer;
    });

    this.node.storage._levelRenderer = this;

    this._setUp = true;
    console.log("    }}} LevelRenderer::setUp");
  }

  tearDown() {
    if (!this._setUp)
      console.error("tearDown called twice!", this);

    console.log("    {{{ LevelRenderer::tearDown", this.node.data);

    this.node.children.forEach((child) => {
      child.storage._nodeRenderer.tearDown();
      child.storage._nodeRenderer = null;
    });

    this.node.storage._levelRenderer = null;

    this._setUp = false;
    console.log("    }}} LevelRenderer::tearDown");
  }

  private makeNodeRenderer(node : SOINode) : NodeRenderer {
    // choose a constructor based on node.data
    if (node.data instanceof Body) {
      var body = node.data as Body;
      return new SphericalNodeRenderer(node, body, this);
    } else {
      console.error("node.data is not Body", node);
    }
  }

  /*
   * Called by the cursor to signal position change.
   * curPos is the position of the cursor relative to the SOINode.
   */
  onUpdatePosition(curPos : Vector3) {
    (this.node.storage._nodeRenderer as NodeRenderer).update(V3(), curPos);

    this.node.children.map((child) => {
      (child.storage._nodeRenderer as NodeRenderer).update(child.relPos, curPos);
    });
  }

  private getNodes() : SOINode[]  {
    var xs = _.clone(this.node.children);
    xs.push(this.node);
    return xs;
  }
}


interface NodeRenderer {
  setUp();
  tearDown();
  update(relPos : Vector3, cursorPos : Vector3);
}


/** A level of detail mesh definition. */
interface LOD {
  /** The minimum angular diameter to activate this LOD. */
  angularDiameter : number;
  /** The level of detail number */
  detail : number;
  /** The mesh at the current LOD. */
  mesh : THREE.Mesh;
}

class SphericalNodeRenderer implements NodeRenderer {
  private _setUp = false;
  private _sceneTargets : THREE.Mesh[];

  public label : Label;
  public meshes : LOD[];
  public currentLOD : number;

  public angularDiameters = [0.5 * deg, 2 * deg,  8 * deg, Infinity];
  public details =          [6,         16,      64,       128];

  public line = null as THREE.Line;
  public updateOrbit : () => void;

  constructor(
    public node : SOINode,
    public body : Body,
    public levelRenderer : LevelRenderer,
    public scene = levelRenderer.scene
  ) {
    this.meshes = [];
    this.currentLOD = -1;
    this.label = LabelManager.instance.createLabel();
    this.label.setText(this.body.name);

    this._sceneTargets = (this.scene as any)._targets as THREE.Mesh[];
  }

  private _ensureOrbitLine() {
    if (this.line) return;

    var body = this.body;
    if (body.kind == "planet" || body.kind == "moon") {
      var orbit = body.orbit.makePlot();
      this.line = orbit.line;
      this.line.name = "Orbit Line: " + this.body.name;
      this.line.position.set(0, 0, 0);
      this.line.scale.set(1, 1, 1);
      this.line.quaternion.setFromAxisAngle(V3(1, 0, 0), 0);

      this.updateOrbit = orbit.updateGeom;
      (this.scene as any)._orbitLines.add(this.line);
      this.updateOrbit();
    }
  }

  private _updatePosition() {
    var orbit = this.body.orbit;
    if (orbit) {
      var t = this.node.soiTree.time;
      this.node.relPos.copy(orbit.getPos(t));
    }

    //if (this.updateOrbit) {
    //  this.updateOrbit();
    //}
  }

  private _updateRotation(mesh : THREE.Mesh) {
    var rotation = this.body.rotation;
    if (rotation) {
      var t = this.node.soiTree.time;
      mesh.quaternion.copy(rotation.getRotation(t));
    }
  }

  /**
   * Update the sphere's position based on this object's displacement to the
   * cursor (displacement is actually position in WebGL world space).
   */
  update(relPos : Vector3, cursorPos : Vector3) {
    if (this.currentLOD == -1) console.error(this.body.name + ": currentLOD == -1", this);
    if (!this._setUp) console.error(this.body.name + ": _setUp is false!", this);

    var displacement = V3().subVectors(relPos, cursorPos);

    var maxDistance = 100 * au, minTheta = this.body.kind == "star" ? 0.006 : 0.003;
    var minThetaRaycastGeom = 0.01;
    var radius = this.body.radius;
    var distance = displacement.length();

    // angular radius on screen in radians
    var theta = radius / distance;

    // Update LOD
    var newLevel = 0;
    while (newLevel < this.meshes.length && this.meshes[newLevel].angularDiameter < 2 * theta) newLevel++;
    if (newLevel >= this.meshes.length) newLevel = this.meshes.length - 1;

    if (newLevel != this.currentLOD) {
      if (this.meshes[newLevel].mesh == null) {
        console.log(this.body.name + ": Requesting LOD " + newLevel, this);
        this.ensureLOD(newLevel);
      } else {
        console.log(
          this.body.name + ": Switch LOD from "
          + this.currentLOD + " to " + newLevel +
          ", angularDiameter:" + (2 * theta * 180 / Math.PI).toFixed(2) + " deg" +
          ", radius:" + (radius * 1e-3).toFixed(2) + " km" +
          ", distance:" + (distance * 1e-3).toFixed(2) + " km", this);
        this.deselect(this.currentLOD);
        this.select(newLevel);
      }
    }

    var mesh = this.meshes[this.currentLOD].mesh;

    if (mesh == null) {
      console.error(this.body.name + ": mesh at " + this.currentLOD + " is null!", this);
      return;
    }

    this._updatePosition();
    this._updateRotation(mesh);

    // Calculate new mesh position
    var actualDistance = distance;

    mesh.position.copy(displacement);

    // Set position of orbit renderer
    // displacement = relPos - cursorPos
    // linePos = displacement - relPos = (relPos - cursorPos) - relPos
    //                                 = -cursorPos
    if (this.line) this.line.position.copy(cursorPos).negate();

    // radius to preserve angular diameter while changing distance
    var scaledRadius = theta * actualDistance;

    // set mesh radius to fit angular diameter
    var size = Math.max(minTheta * actualDistance, scaledRadius);
    mesh.scale.set(size, size, size);

    // set raycast geometry scale
    var raycastGeom : THREE.Mesh = (mesh as any)._raycastGeometry;

    var objTheta = size / actualDistance;
    var raycastGeomScale = objTheta > minThetaRaycastGeom ? 1 : minThetaRaycastGeom / objTheta;
    raycastGeom.scale.set(raycastGeomScale, raycastGeomScale, raycastGeomScale);

    var light : THREE.DirectionalLight = (mesh as any)._light;
    if (light) {
      light.lookAt(V3());
    }

    mesh.updateMatrix();
    mesh.updateMatrixWorld(false);

    this.label.reposition(mesh);

    var otherPos = V3().copy(this.levelRenderer.treeRenderer.cursor.localPos);
    var relPos = this.levelRenderer.treeRenderer.cursor.cursor.relativePositionTo(this.node);

    var dist = otherPos.add(relPos).length();
    this.label.setText(this.body.name); // + " (" + formatDistance(dist, 2) + ")");
  }

  setUp() {
    this._setUp = true;
    this.meshes = _.zipWith(this.angularDiameters, this.details, (angularDiameter, detail) => {
      return {angularDiameter: angularDiameter as number, detail: detail as number, mesh: null} as LOD;
    }) as LOD[];

    this._ensureOrbitLine();
    this.ensureLOD(0);
    this.select(0);
  }

  tearDown() {
    this.deselect(this.currentLOD);
    for (var i = 0; i < this.meshes.length; i++) this.removeLOD(i);
    this._setUp = false;
    this.label.remove();

    if (this.line) {
      (this.scene as any)._orbitLines.remove(this.line);
      this.updateOrbit = null;
      this.line = null;
    }
  }

  private ensureLOD(index : number) {
    if (this.meshes[index].mesh != null) return;
    var lod = this.meshes[index];
    var detail = lod.detail;

    var geometry = new THREE.SphereGeometry(1, detail, detail);
    var material : THREE.Material;

    if (detail < 10 || this.body.kind == "star") {
      var farColor = Col().copy(this.body.color).offsetHSL(0, -0.1, 0.1);
      material = new THREE.MeshBasicMaterial({
        color: farColor.getHex()
      });
    } else {
      if (this.body.material != null)
        material = this.body.material;
      else
        material = new THREE.MeshPhongMaterial({
          color: this.body.color.getHex(),
          specular: 0x222222,
          shininess: 20
        });
    }

    var mesh = new THREE.Mesh(geometry, material);

    if (this.body.data.rings && detail > 10) {
      var ringsMesh = MakeRingMesh(this.body.data.rings);
      mesh.add(ringsMesh);

      ringsMesh.name = this.body.name + ": Rings";
      ringsMesh.position.set(0, 0, 0);
      var scale = 1 / this.body.data.radius;
      ringsMesh.scale.set(scale, scale, scale);
    }

    mesh.name = "SphericalNodeRenderer: " + this.body.toString();
    (mesh as any)._renderer = this;
    lod.mesh = mesh;

    if (this.body.kind == "star") {
      // For far distances, the sun is a directional light, but in this
      // game, there are no situations where the arrangement bodies where
      // the directional light approximation breaks down and point light is
      // needed. (things behind the sun have low LOD so they are not lighted)
      var light = new THREE.DirectionalLight(0xffffff, 0.8);
      mesh.add(light);
      (mesh as any)._light = light;
    }

    var raycastGeom = new THREE.Mesh(
      new THREE.SphereGeometry(2, 6, 6),
      new THREE.MeshBasicMaterial({
        color: 0xffffff, opacity: 0, transparent: true, depthWrite: false
      }));
    raycastGeom.name = this.body.name + ": raycastGeom";

    (mesh as any)._raycastGeometry = raycastGeom;
    (raycastGeom as any)._sphericalNodeRenderer = this;

    mesh.add(raycastGeom);

    this._sceneTargets.push(raycastGeom);

    console.log(
      this.body.name + ": ensureLOD: LOD instantiated: " +
      index + " (detail=" + detail + ")", lod, this);
  }

  private removeLOD(index : number) {
    var lod = this.meshes[index];

    if (lod && lod.mesh) {
      var raycastGeom = (lod.mesh as any)._raycastGeometry;
      _.remove(this._sceneTargets, (t) => t === raycastGeom);
    }

    lod.mesh = null;
    this.meshes[index] = null;
  }

  private select(index : number) {
    if (index == -1) return;
    this.scene.add(this.meshes[index].mesh);
    this.currentLOD = index;
  }

  private deselect(index : number) {
    if (index == -1) return;
    var mesh = this.meshes[index].mesh;
    (mesh as any)._renderer = null;
    this.scene.remove(mesh);
    this.currentLOD = -1;
  }
}


class SOITreeRenderer {
  public renderers : LevelRenderer[];
  constructor(
    public tree : SOITree,
    public scene : THREE.Scene,
    public cursor = new SOICursor(tree, tree.root)
  ) {

    this.cursor.onMoveToChild((ch) => {
      this.pushRenderer(ch);
    });

    this.cursor.onMoveToParent((p) => {
      this.popRenderer();
    });

    this.renderers = [];
    this.pushRenderer(this.tree.root);
  }

  private pushRenderer(node : SOINode) {
    console.log("{{{ pushRenderer", node.data);
    var r = this.getLevel(node);
    r.setUp();
    this.renderers.push(r);
    console.log("}}} pushRenderer", this.renderers);
  }

  private popRenderer() {
    console.log("{{{ popRenderer");
    var r = this.getTopRenderer();
    r.tearDown();
    this.renderers.pop();
    console.log("}}} popRenderer", this.renderers);
  }

  private getTopRenderer() {
    return this.renderers[this.renderers.length - 1];
  }

  private getLevel(node : SOINode) {
      if (node.storage._levelRenderer)
        return node.storage._levelRenderer;
      else
        return new LevelRenderer(node, this);
  }

  moveCursor(delta : Vector3) {
    this.cursor.move(delta);

    // relative position of current node, which computed by doing is partial sum
    // from children to root
    var curRelPos = V3().copy(this.cursor.localPos);

    _(this.renderers).forEachRight((renderer_) => {
      var renderer = renderer_ as LevelRenderer;
      renderer.onUpdatePosition(V3().copy(curRelPos));
      curRelPos.add(renderer.node.relPos);
    });
  }

  getRelativePositionTo(node : SOINode) {
    return this.cursor.cursor.relativePositionTo(node).add(this.cursor.localPos);
  }
}


function getSolarSystem() : SOITree {
  var solarSystemData = getSolarSystemData();
  var solarSystemPairs = (_.toPairs(solarSystemData) as any) as [{0: string, 1: BodyData}];

  var solarSystemRadius = 1.2 * _(solarSystemPairs).map(
    (kv) => kv[1].orbit.semiMajorAxis).max();

  var sun = new Body("Sun", {radius: 695500 * km, color: 0xffff00}, "star");
  var solarSystem = new SOITree(solarSystemRadius, sun);

  var solarSystemItems = {};

  var planets = _.filter(solarSystemPairs, (kv) => kv[1].parent == null);
  var moons = _.filter(solarSystemPairs, (kv) => kv[1].parent != null);

  [planets, moons].forEach((arr) => {
    arr.forEach((kv) => {
      var name = kv[0], data = kv[1];

      var isMoon = data.parent != null;
      var color = data.color;
      var body = new Body(name, data, isMoon ? "moon" : "planet");
      var soiRadius = isMoon ? data.radius * 25 : data.radius * 3500;

      var theta = data.orbit.phase,  R = data.orbit.semiMajorAxis;
      var pos = V3(Math.cos(theta) * R, 0, Math.sin(theta) * R);

      var parent = (isMoon ? solarSystemItems[data.parent] : solarSystem.root);
      var node = parent.addChild(pos, soiRadius, body);

      solarSystemItems[name] = node;
    });
  });

  function testRelativePositionTo() {
    var moon = (solarSystemItems as any).Moon;
    var earth = (solarSystemItems as any).Earth;
    var mars = (solarSystemItems as any).Mars;
    var jupiter = (solarSystemItems as any).Jupiter;
    var saturn = (solarSystemItems as any).Saturn;

    console.log(moon.getRelativePositionList());

    // should be zero vector
    console.log("moon/moon", moon.relativePositionTo(moon));

    // should be negation of earth other
    console.log("moon/earth", moon.relativePositionTo(earth),
                "earth/moon", earth.relativePositionTo(moon));
    console.log("moon/sun", moon.relativePositionTo(solarSystem.root),
                "sun/moon", solarSystem.root.relativePositionTo(moon));

    // chain addition property of relative positions
    var moonEarth = moon.relativePositionTo(earth);
    var earthMars = earth.relativePositionTo(mars);
    var marsJupiter = mars.relativePositionTo(jupiter);
    var jupiterSaturn = jupiter.relativePositionTo(saturn);
    var sum = V3().add(moonEarth).add(earthMars).add(marsJupiter).add(jupiterSaturn);

    console.log("moon/saturn", moon.relativePositionTo(saturn),
                "= moon/earth", moonEarth,
                "+ earth/mars", earthMars,
                "+ mars/jupiter", marsJupiter,
                "+ jupiter/saturn", jupiterSaturn,
                "= moon/saturn", sum);
  }

  //testRelativePositionTo();

  return solarSystem;
}

