/// <reference path="three.d.ts" />
/// <reference path="Utils.ts" />

// OrbitalMechanics.ts

// Adapted from P2 code.

/** Plot function f : R -> R^3 */
function MakePlot(
  f : (t : number, v : Vector3) => void,
  mat : THREE.LineDashedMaterial | THREE.LineBasicMaterial | THREE.ShaderMaterial,
  params : {closed? : boolean, t0? : number, t1? : number, dt? : number}) {
  'use strict';
  if (!params) params = {};

  var closed = params.closed || false;
  var t0 = params.t0 || 0.0, t1 = params.t1 || 1.0, dt = params.dt || 1e-2;

  var geom = new THREE.Geometry();

  geom.vertices = [];
  var k = 0;
  for (var t = t0; t <= t1; t += dt) {
    geom.vertices.push(V3());
    k++;
  }
  if (closed) geom.vertices.push(V3());

  function updateGeom() {

    var t = t0;
    for (var i = 0; i < k; i++) {
      f(t, geom.vertices[i]);
      t += dt;
    }

    if (closed)
      geom.vertices[k].copy(geom.vertices[0]);

    geom.computeLineDistances();
    geom.verticesNeedUpdate = true;
  }

  var line = new THREE.Line(geom, mat);
  line.renderOrder = -100;
  return {line: line, updateGeom: updateGeom};
}


interface OrbitData {
  semiMajorAxis : number;
  eccentricity? : number;
  period : number;
  phase? : number;
  inclination? : number;
  argumentOfPeriapsis? : number;
  longitudeOfAscendingNode? : number;
}


interface RotationData {
  period : number;
  tilt? : number;
  phase? : number;
}


/** Handles computations related to orbits */
class Orbit {
  private _quat = new THREE.Quaternion();

  constructor(public params : OrbitData) {
    this._updateQuaternion();
  }

  makePlot() {
    var vec = V3();
    var mat = new THREE.LineBasicMaterial({
      linewidth: 0.4, color: 0xffffff,
      opacity: 0.3, transparent: true
    });
    this._updateQuaternion();
    return MakePlot(
      (t, v) => { return this.getPos(t, true, v, null); },
        mat, {
        t0: 0, t1: 2*Math.PI,
        dt: 2*Math.PI / 2160, closed : this.params.eccentricity < 1 });
  }

  private _updateQuaternion() {
    // Get the quaternion for orbital plane given
    //   inclination (i)
    //   longitude of ascending node (Omega)
    //   arg of periapsis (w)

    var qw = quatAxisAngle(V3(0, 1, 0), this.params.argumentOfPeriapsis || 0.0);
    var qi = quatAxisAngle(V3(1, 0, 0), this.params.inclination || 0.0);
    var qOmega = quatAxisAngle(V3(0, 1, 0), this.params.longitudeOfAscendingNode || 0.0);

    this._quat.copy(qOmega.multiply(qi.multiply(qw)));
  }

  /** Get the quaternion for the orbital plane. */
  public getOrbitalQuaternion() {
    //this._updateQuaternion();
    return new THREE.Quaternion().copy(this._quat);
  }

  /**
   * Get position relative to ecliptic.
   */
  getPos(t : number, asTheta = false, into : Vector3 = null, getVel = null) {
    var vec = this.getInPlanePos(t, asTheta, into, getVel);

    //this._updateQuaternion();
    if (getVel) getVel.applyQuaternion(this._quat);
    return vec.applyQuaternion(this._quat);
  }

  getVel(t : number, asTheta = false) {
    var vel = V3();
    //this._updateQuaternion();
    this.getInPlanePos(t, asTheta, null, vel);
    return vel.applyQuaternion(this._quat);
  }

  /**
    * Get position relative to orbital plane at any given time.
    * If asTheta is false, then t is the current time, otherwise, t is true
    * anomaly.
    */
  getInPlanePos(t : number, asTheta = false, into = null as Vector3, getVel = null as Vector3) {
    'use strict';
    var params = this.params;

    var e = params.eccentricity || 0.0;
    var T = params.period || 1;
    var phi = params.phase || 0;
    var sMa = params.semiMajorAxis || 0.0;
    var p = sMa * (1 - e*e);

    var M, E, nu;
    if (asTheta) {
      M = E = nu = t;
    } else {
      // Mean anomaly
      M = (phi + 2 * Math.PI * (t / T)) % (2 * Math.PI);

      // Find eccentric anomaly E from Kepler's Equation
      // https://en.wikipedia.org/wiki/Kepler%27s_equation#Inverse_problem
      // Initial guess using 2nd order series expansion
      var _e = 1 - e;
      E = (M / _e) * (1 - M * M * e / (6 * _e*_e*_e));

      // Use Newton's method
      var delta = 1000, iters = 0;
      // It takes < 10 iterations for e < 0.1
      while (iters < 1000) {
        if (Math.abs(delta) < 1e-3) break;

        // E1 = E0 - f(E0) / f'(E0)
        E = E - (delta = E - e * Math.sin(E) - M) /
          (1 - e * Math.cos(E));
        iters++;
      }

      // Calculate true anomaly
      var argx = Math.sqrt(1 + e) * Math.sin(E / 2);
      var argy = Math.sqrt(1 - e) * Math.cos(E / 2);
      nu = 2 * Math.atan2(argy, argx);
    }

    var r = p / (1 + e * Math.cos(Math.PI + nu));

    var x = r * Math.sin(nu), y = -r * Math.cos(nu);

    if (getVel) {
      var dx = Math.cos(nu), dy = Math.sin(nu);

      // https://en.wikipedia.org/wiki/Standard_gravitational_parameter
      var mu = 4 * (Math.PI * Math.PI) * (sMa * sMa * sMa) / (T * T);
      // https://en.wikipedia.org/wiki/Orbital_speed
      var v = Math.sqrt(mu * (2 / r - 1 / sMa));

      getVel.set(v * dx, 0, v * dy);
    }

    if (into) return into.set(x, 0, y);
    else return V3(x, 0, y);
  };
}

class Rotation {
  constructor(public params : RotationData) { }

  public getRotation(t) {
    function _getRotQuaternion(omega, i, theta) {
      var qtheta = quatAxisAngle(new THREE.Vector3(0, 1, 0), theta);
      var qi = quatAxisAngle(new THREE.Vector3(0, 0, -1), i);
      var qomega = quatAxisAngle(new THREE.Vector3(0, 1, 0), omega);
      return qomega.multiply(qi.multiply(qtheta));
    }

    var params : RotationData = this.params;
    var phi = params.phase || 0.0;
    var theta = params.period ? 2 * Math.PI * (t / params.period) : 0;

    var q = _getRotQuaternion(0, params.tilt || 0.0, phi + theta);
    return q;
  }
}
